#include "MultiReader.h"

#include <sys/poll.h>	// poll
#include <unistd.h>		// fcntl
#include <fcntl.h>		// fcntl
#include <errno.h>

#include <Logger.h>

#undef LOGTAG
#define LOGTAG "MultiReader"

using namespace std;
using namespace stringf;

void MultiReader::addFD(int fd, Listener* l) throw ()
{
	listeners.insert(pair<int, Listener*>(fd, l));
}

bool MultiReader::readLoop() throw (SystemException)
{
	// Initialize the poll data structures
	int pollcount = listeners.size();
	struct pollfd p[pollcount];
	int c = 0;
	for (map<int, Listener*>::const_iterator i = listeners.begin();
			i != listeners.end(); i++, c++)
	{
		p[c].fd = i->first;
		p[c].events = POLLIN;
		p[c].revents = 0;
	}

	// While there is output, read it
	bool done = false;
	while (pollcount && !done)
	{
		int res = poll(p, pollcount, -1);
		if (res == -1)
		{
			if (errno != EINTR)
				throw SystemException(errno, "Checking for output on " +
						fmt(pollcount) + " file descriptors");
		} else if (res == 0)
			log_warn("poll returned 0 with an infinite timeout");
		else
		{
			for (int i = 0; i < pollcount; i++)
			{
				if (p[i].revents & POLLIN)
				{
					//fprintf(stderr, "Data available on %d\n", i);
					if (!listeners[p[i].fd]->haveData(p[i].fd))
						done = true;
				}

				if (p[i].revents & POLLHUP)
				{
					//fprintf(stderr, "Closed connection on %d\n", i);
					if (!listeners[p[i].fd]->haveEOF(p[i].fd))
						done = true;

					// Delete the closed handle from the poll list
					--pollcount;
					if (i != pollcount)
						p[i] = p[pollcount];
				}
			}
		}
	}
	return pollcount == 0;
}


bool MultiReader::LineListener::haveData(int fd) throw ()
{
	try {
		int oldfl;
		if (fcntl(fd, F_GETFL, &oldfl) == -1)
			throw SystemException(errno, "Reading file descriptor flags");

		int newfl = oldfl | O_NONBLOCK;
		if (fcntl(fd, F_SETFL, &newfl) == -1)
			throw SystemException(errno, "Setting O_NONBLOCK on file descriptor flags");

		const int bufsize = 512;
		char buf[bufsize];
		int sz;
		while (1)
		{
			sz = read(fd, buf, bufsize);
			if (sz == -1)
				if (errno == EAGAIN)
					break;
				else
					throw SystemException(errno, "Reading from file descriptor");
			if (sz == 0)
				break;
			for (int i = 0; i < sz; i++)
				if (buf[i] == '\n')
				{
					if (!haveLine(line))
						return false;
					line = string();
				} else
					line += buf[i];
		}

		if (fcntl(fd, F_SETFL, &oldfl) == -1)
			throw SystemException(errno, "Restoring old file descriptor flags");

		return true;
	} catch (Exception& e) {
		log_err(string(e.type()) + ": " + e.desc());
		return false;
	}
}

// vim:set ts=4 sw=4:
